from enum import Enum
from typing import Optional, List

from pytos2.models import Jsonable
from pytos2.utils import prop, propify


@propify
class TopologyDevice(Jsonable):
    """
    Represents a topology device in SecureTrack.

    This model maps to the device objects found in the TopologyDevices API response.
    """

    # Standard device properties
    id: int = prop(0, key="device_id")
    name: str = prop("", key="device_name")
    model: str = prop("", key="device_model")
    domain: int = prop(0, key="device_domain")
    vendor: str = prop("", key="device_vendor")
    virtual_type: str = prop("", key="device_virtual_type")
    topology_enabled: bool = prop(False, key="device_topology_enabled")
    has_dynamic_topology: str = prop("disable", key="device_has_dynamic_topology")

    # Optional device properties
    parent: Optional[int] = prop(None, key="device_parent")
    context_name: Optional[str] = prop(None, key="device_context_name")
    firewall_id: Optional[str] = prop(None, key="device_firewall_id")

    # OPM-related properties
    opm_id: Optional[str] = prop(None, key="opm_device_id")
    opm_system_id: Optional[str] = prop(None, key="opm_system_id")
    opm_type: Optional[str] = prop(None, key="opm_type")
    opm_parent: Optional[str] = prop(None, key="opm_parent")
    opm_vendor: Optional[str] = prop(None, key="opm_vendor")
    opm_model: Optional[str] = prop(None, key="opm_model")



import pytest
import json
import responses
from responses import matchers

from pytos2.securetrack.topology_device import TopologyDevice
from pytos2.utils import get_api_node


class TestTopologyDevice:
    @responses.activate
    def test_get_topology_devices(self, st, topology_device_mock):
        """Test getting all topology devices"""
        devices = st.get_topology_devices()

        assert len(devices) == 229
        assert isinstance(devices[0], TopologyDevice)
        assert devices[0].id == 716
        assert devices[0].name == "V_OPM"
        assert devices[0].topology_enabled is True

        hub_dev = next(d for d in devices if d.name == "VirtualHub_West_US")
        assert hub_dev.id == 608
        assert hub_dev.vendor == "Azure"
        assert hub_dev.opm_id == "W-KeQN48tL-c2PXq2TuiYg=="
        assert hub_dev.opm_parent == "r2k5sreJHbdafBDqIQnACw=="
        assert hub_dev.parent == 606

        # Check that different types of devices are parsed correctly
        # Regular device without OPM properties
        asa_device = next(d for d in devices if d.name == "ASA_L2LFW")
        assert asa_device.id == 392
        assert asa_device.vendor == "Cisco"
        assert asa_device.opm_id is None

        # Azure device with OPM properties
        azure_device = next(
            d for d in devices if d.name == "Azure Testing Subscription"
        )
        assert azure_device.id == 508
        assert azure_device.vendor == "Azure"
        assert azure_device.opm_vendor == "MICROSOFT"
        assert azure_device.opm_model == "AZURE_ACCOUNT"
        assert azure_device.opm_system_id == "t7qaK8gDSbC-5tCF9DXy7Q=="

    @responses.activate
    def test_get_topology_device(self, st, topology_device_mock):
        """Test getting a single topology device by ID"""
        device = st.get_topology_device(716)

        assert isinstance(device, TopologyDevice)
        assert device.id == 716
        assert device.name == "V_OPM"
        assert device.model == ""
        assert device.domain == 1
        assert device.virtual_type == ""
        assert device.topology_enabled is True
        assert device.has_dynamic_topology == "enable"
        assert device.opm_id == "aNu2ebd5mhuDsBfm6v1Kog=="
        assert device.opm_system_id == "bvnfqz9aQq-qyAgdfJThVA=="
        assert device.opm_type == "MANAGEMENT"
        assert device.opm_vendor == "UNKNOWN"
        assert device.opm_model == "UNKNOWN"

    @responses.activate
    def test_get_topology_device_not_found(self, st, topology_device_mock):
        """Test getting a topology device that doesn't exist"""
        with pytest.raises(ValueError, match="Error Getting Topology Device 999"):
            st.get_topology_device(999)

    @responses.activate
    def test_topology_device_properties(self, topology_devices_data):
        """Test the TopologyDevice model properties and conversion"""
        device_data = topology_devices_data["TopologyDevices"][0]
        device = TopologyDevice.kwargify(device_data)

        assert device.id == 716
        assert device.name == "V_OPM"
        assert device.domain == 1
        assert device.topology_enabled is True
        assert device.opm_id == "aNu2ebd5mhuDsBfm6v1Kog=="
        assert device.opm_type == "MANAGEMENT"

        # Test conversion back to JSON/dict representation
        device_dict = device._json
        assert device_dict["device_id"] == 716
        assert device_dict["device_name"] == "V_OPM"
        assert device_dict["device_topology_enabled"] is True
        assert device_dict["opm_device_id"] == "aNu2ebd5mhuDsBfm6v1Kog=="



from datetime import date, datetime
from types import EllipsisType
from typing import Union, Optional, List

from netaddr import IPAddress
from requests import Response
from requests.exceptions import HTTPError

# avoid circular imports
import pytos2
from pytos2.api import ApiError, SdkError
from pytos2.models import DateFilterType, coerce_date_filter_type
from pytos2.secureapp.access_requests import ApplicationAccessRequest
from pytos2.securechange.ticket import Ticket  # noqa
from .api import SaAPI
from pytos2.utils import NoInstance, get_api_node
from .application_identities import ApplicationIdentity
from .applications import (
    Application,
    ApplicationConnection,
    ApplicationInterface,
    ApplicationInterfaceInstance,
    ApplicationInterfaceInstanceGet,
    Connection,
    ExtendedApplicationConnection,
    InterfaceConnection,
)
from .network_object import NetworkObject
from .service_object import Service
from .history import HistoryRecord
from pytos2.securechange.user import SCWUser
from pytos2.securechange.fields import Field
from pytos2.models import ObjectReference


class Sa:
    default: Union["Sa", NoInstance] = NoInstance(
        "Sa.default",
        "No Sa instance has been initialized yet, initialize with `Sa(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
    ):
        self.api: SaAPI = SaAPI(hostname, username, password)
        if default:
            Sa.default = self

        self._application_identities: List[ApplicationIdentity] = []

    @property
    def application_identities(self) -> List[ApplicationIdentity]:
        if not self._application_identities:
            res = self._get_application_identities()
            self._application_identities = res
        return self._application_identities

    def _get_application_identities(self, cache=True) -> List[ApplicationIdentity]:
        res = self.api.get_application_identities()
        data = self.api.handle_json(res, "application_identities")
        return [
            ApplicationIdentity.kwargify(a)
            for a in get_api_node(
                data, "application_identities.application_identity", listify=True
            )
        ]

    def get_applications(self) -> List[Application]:
        """
        This function returns a list of all applications in SecureApp.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications

        Usage:
            apps = sa.get_applications()
        """

        res = self.api.get_applications()
        data = self.api.handle_json(res, "get_applications")
        return [
            Application.kwargify(a)
            for a in get_api_node(data, "applications.application", listify=True)
        ]

    def get_application(self, application_id: str) -> Application:
        """
        This function returns a single application by its ID.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            app = sa.get_application(60)
        """

        res = self.api.get_application(application_id)
        data = self.api.handle_json(res, "get_application")
        application = get_api_node(data, "application")
        if isinstance(application, list) and application:
            application = application[0]

        return Application.kwargify(application)

    def add_application(
        self,
        application: Optional[Application] = None,
        /,
        *,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        owner: Union[None, SCWUser, int, str] = None,
        editors: Optional[List[Union[SCWUser, int]]] = None,
        viewers: Optional[List[Union[SCWUser, int]]] = None,
        customer: Optional[
            int
        ] = None,  # TODO: Add customer to this once it's implemented.
    ):
        """
        This function creates a new application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications

        Usage:
            Example 1:
            app = sa.add_application(name="My App", comment="This is my app")

            Example 2:
            app = sa.add_application(
                name="My App",
                owner=100,
                editors=[101, 102],
                viewers=[103, 104]
            )
        """

        from pytos2.securechange import Scw

        if not application:
            if isinstance(owner, str):
                owner = Scw.default.get_user(owner)
            if owner:
                owner = owner if isinstance(owner, int) else owner.id

            editors = editors or []
            viewers = viewers or []

            editors = [e if isinstance(e, int) else e.id for e in editors]
            viewers = [v if isinstance(v, int) else v.id for v in viewers]

            application = Application(
                name=name,
                comment=comment,
                owner=ObjectReference(id=owner) if owner else None,
                editors=[ObjectReference(id=e.id) for e in editors],
                viewers=[ObjectReference(id=v.id) for v in viewers],
                customer=ObjectReference(id=customer) if customer else None,
            )

        res = self.api.add_application(application._json)
        return self.api.handle_creation(res, "add_application", cls=Application)

    def update_application(
        self,
        application_id: Union[int, Application],
        name: Optional[str] = None,
        comment: Optional[str] = None,
        owner: Union[None, SCWUser, int] = None,
        editors: Optional[List[Union[SCWUser, int]]] = None,
        viewers: Optional[List[Union[SCWUser, int]]] = None,
    ):
        """
        This function updates an existing application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            Example 1:
            sa.update_application(60, name="My App", comment="This is my app")

            Example 2:
            sa.update_application(
                60,
                name="My App",
                owner=100,
                editors=[101, 102],
                viewers=[103, 104]
            )
        """
        if owner:
            owner = owner if isinstance(owner, int) else owner.id

        editors = (
            [e if isinstance(e, int) else e.id for e in editors] if editors else None
        )
        viewers = (
            [v if isinstance(v, int) else v.id for v in viewers] if viewers else None
        )

        if isinstance(application_id, Application):
            app = application_id
            application_id = app.id
        else:
            app = Application(
                name=name,
                comment=comment,
                owner=ObjectReference(id=owner) if owner else None,
                editors=(
                    [ObjectReference(id=e.id) for e in editors] if editors else None
                ),
                viewers=(
                    [ObjectReference(id=v.id) for v in viewers] if viewers else None
                ),
            )

        res = self.api.update_application(application_id, app._json)
        res = self.api.handle_response(res, "update_application", action="update")

    def delete_application(self, application_id: int):
        """
        This function deletes an application by its ID.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/60

        Usage:
            sa.delete_application(60)
        """

        res = self.api.delete_application(application_id)
        self.api.handle_response(res, "delete_application", action="delete")

    def bulk_update_applications(self, apps: List[Application]):
        res = self.api.update_applications([a._json for a in apps])
        self.api.handle_response(res, "bulk_update_applications", action="update")

    def get_application_connections(
        self, application_id: int
    ) -> List[ApplicationConnection]:
        """
        This function returns a list of all connections for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections

        Usage:
            connections = sa.get_application_connections(60)
        """

        res = self.api.get_application_connections(application_id)
        data = self.api.handle_json(res, "get_application_connections")
        connections = get_api_node(data, "connections.connection", listify=True)
        return [ApplicationConnection.kwargify(c) for c in connections]

    def get_application_connection(
        self, application_id: int, connection_id: int
    ) -> ApplicationConnection:
        """
        This function returns a single connection for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections/100

        Usage:
            connection = sa.get_application_connection(60, 100)
        """

        res = self.api.get_application_connection(application_id, connection_id)
        data = self.api.handle_json(res, "get_application_connection")
        conn = get_api_node(data, "connection")
        return ApplicationConnection.kwargify(conn)

    def get_extended_application_connections(
        self, application_id: int
    ) -> List[ApplicationConnection]:
        """
        This function returns a list of all extended connections for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_extended

        Usage:
            connections = sa.get_extended_application_connections(60)
        """

        res = self.api.get_extended_application_connections(application_id)
        data = self.api.handle_json(res, "get_extended_application_connections")
        connections = get_api_node(
            data, "connections_extended.connection_extended", listify=True
        )
        return [ExtendedApplicationConnection.kwargify(c) for c in connections]

    def add_application_connection(
        self,
        application_id: int,
        connection: Optional[ApplicationConnection] = None,
        /,
        *,
        services: Optional[List[Union[Service, ObjectReference, int]]] = None,
        sources: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        destinations: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        status: Optional[str] = None,
        external: Optional[bool] = None,
        connection_to_application: Union[None, ObjectReference, int] = None,
        comment: Optional[str] = None,
        uid: Optional[str] = None,
        open_tickets: Optional[List[Union[ObjectReference, int]]] = None,
        name: Optional[str] = None,
    ):
        """
        This function creates a new connection for a given application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections

        Usage:
            Example 1:
            sa.add_application_connection(
                60,
                services=[Service(id=1), Service(id=2)],
                sources=[NetworkObject(id=1), NetworkObject(id=2)],
                destinations=[NetworkObject(id=3), NetworkObject(id=4)],
            )

            Example 2:
            sa.add_application_connection(
                60,
                connection=ApplicationConnection(
                    services=[Service(id=1), Service(id=2)],
                    sources=[NetworkObject(id=1), NetworkObject(id=2)],
                    destinations=[NetworkObject(id=3), NetworkObject(id=4)],
                )
            )
        """

        services = (
            [s if isinstance(s, int) else s.id for s in services] if services else []
        )
        sources = (
            [s if isinstance(s, int) else s.id for s in sources] if sources else []
        )
        destinations = (
            [d if isinstance(d, int) else d.id for d in destinations]
            if destinations
            else []
        )
        connection_to_application = (
            (
                connection_to_application.id
                if isinstance(connection_to_application, ObjectReference)
                else connection_to_application
            )
            if connection_to_application
            else []
        )
        open_tickets = (
            [t if isinstance(t, int) else t.id for t in open_tickets]
            if open_tickets
            else []
        )

        if not connection:
            connection = ApplicationConnection(
                services=[ObjectReference(id=s) for s in services],
                sources=[ObjectReference(id=s) for s in sources],
                destinations=[ObjectReference(id=d) for d in destinations],
                status=status,
                external=external,
                connection_to_application=ObjectReference(id=connection_to_application),
                comment=comment,
                uid=uid,
                open_tickets=[ObjectReference(id=t) for t in open_tickets],
                name=name,
            )
        res = self.api.add_application_connection(application_id, connection._json)
        return self.api.handle_creation(
            res, "add_application_connection", cls=ApplicationConnection
        )

    def update_application_connection(
        self,
        application_id: int,
        connection_id: Union[int, ApplicationConnection],
        /,
        *,
        services: Optional[List[Union[Service, ObjectReference, int]]] = None,
        sources: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        destinations: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        status: Optional[str] = None,
        external: Optional[bool] = None,
        connection_to_application: Union[None, ObjectReference, int] = None,
        comment: Optional[str] = None,
        uid: Optional[str] = None,
        open_tickets: Optional[List[Union[ObjectReference, int]]] = None,
        name: Optional[str] = None,
    ):
        """
        This function updates an existing connection for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections/100

        Usage:
            Example 1:
            sa.update_application_connection(
                60,
                100,
                services=[Service(id=1), Service(id=2)],
                sources=[NetworkObject(id=1), NetworkObject(id=2)],
                destinations=[NetworkObject(id=3), NetworkObject(id=4)],
            )

            Example 2:
            sa.update_application_connection(
                60,
                ApplicationConnection(
                    id=100,
                    services=[Service(id=1), Service(id=2)],
                    sources=[NetworkObject(id=1), NetworkObject(id=2)],
                    destinations=[NetworkObject(id=3), NetworkObject(id=4)],
                )
            )
        """

        if isinstance(connection_id, ApplicationConnection):
            conn = connection_id
            connection_id = conn.id
        else:
            conn = ApplicationConnection(
                services=services,
                sources=sources,
                destinations=destinations,
                status=status,
                external=external,
                connection_to_application=connection_to_application,
                comment=comment,
                uid=uid,
                open_tickets=open_tickets,
                name=name,
            )

        res = self.api.update_application_connection(
            application_id, connection_id, conn._json
        )
        self.api.handle_response(res, "update_application_connection")

    def bulk_update_application_connections(
        self, application_id: int, connections: List[ApplicationConnection]
    ):
        """
        This function updates multiple connections for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections

        Usage:
            sa.update_application_connections(60, [ApplicationConnection(id=100, status="ACTIVE")])
        """
        res = self.api.update_application_connections(
            application_id, [c._json for c in connections]
        )
        self.api.handle_response(
            res, "bulk_update_application_connections", action="update"
        )

    def delete_application_connection(self, application_id: int, connection_id: int):
        """
        This function deletes a connection for a given application.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections/100

        Usage:
            sa.delete_application_connection(60, 100)
        """

        res = self.api.delete_application_connection(application_id, connection_id)
        self.api.handle_response(res, "delete_application_connection", action="delete")

    def get_application_interfaces(self, application_id: int):
        """
        This function returns a list of all interfaces for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces

        Usage:
            interfaces = sa.get_application_interfaces(60)

        """

        res = self.api.get_application_interfaces(application_id)
        data = self.api.handle_json(res, "get_application_interfaces")
        interfaces = get_api_node(
            data, "application_interfaces.application_interface", listify=True
        )
        return [ApplicationInterface.kwargify(i) for i in interfaces]

    def get_application_interface(self, application_id: int, interface_id: int):
        """
        This function returns a single interface for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/242/interfaces/31

        Usage:
            interface = sa.get_application_interface(242, 31)
        """

        res = self.api.get_application_interface(application_id, interface_id)
        data = self.api.handle_json(res, "get_application_interface")
        data = get_api_node(data, "application_interface")

        if isinstance(data, list) and data:
            data = data[0]

        return ApplicationInterface.kwargify(data)

    def add_application_interface(
        self,
        application_id: int,
        interface: Optional[ApplicationInterface] = None,
        /,
        *,
        name: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        """
        This function creates a new interface for a given application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces

        Usage:
            Example 1:
            sa.add_application_interface(
                60,
                name="My Interface",
                comment="This is my interface",
            )

            Example 2:
            sa.add_application_interface(
                60,
                ApplicationInterface(
                    name="My Interface",
                    comment="This is my interface",
                )
            )
        """

        interface = (
            ApplicationInterface(
                name=name,
                comment=comment,
            )
            if interface is None
            else interface
        )

        res = self.api.add_application_interface(application_id, interface._json)
        return self.api.handle_creation(
            res, "add_application_interface", cls=ApplicationInterface
        )

    def update_application_interface(
        self,
        application_id: int,
        interface_id: int | ApplicationInterface,
        /,
        *,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        is_published: Optional[bool] = None,
    ):
        """
        This function updates an existing interface for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces/31

        Usage:
            Example 1:
            sa.update_application_interface(
                60,
                31,
                name="My Interface",
                comment="This is my interface",
                is_published=True,
            )

            Example 2:
            sa.update_application_interface(
                60,
                ApplicationInterface(
                    id=31,
                    name="My Interface",
                    comment="This is my interface",
                    is_published=True,
                )
            )
        """

        if isinstance(interface_id, ApplicationInterface):
            interface = interface_id
            interface_id = interface_id.id
        else:
            interface = ApplicationInterface(
                name=name,
                comment=comment,
                is_published=is_published,
            )

        res = self.api.update_application_interface(
            application_id, interface_id, interface._json
        )
        self.api.handle_response(res, "update_application_interface", action="update")

    def delete_application_interface(self, application_id: int, interface_id: int):
        """
        This function deletes an interface for a given application.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/242/interfaces/31

        Usage:
            sa.delete_application_interface(242, 31)
        """

        res = self.api.delete_application_interface(application_id, interface_id)
        self.api.handle_response(res, "delete_application_interface", action="delete")

    def get_application_interface_connections(
        self, application_id: int, interface_id: int
    ):
        """
        This function returns a list of all connections for a given interface and application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces/31/connections

        Usage:
            connections = sa.get_application_interface_connections(60, 31)
        """

        res = self.api.get_application_interface_connections(
            application_id, interface_id
        )
        data = self.api.handle_json(res, "get_application_interface_connections")
        connections = get_api_node(
            data, "interface_connections.interface_connection", listify=True
        )
        return [InterfaceConnection.kwargify(c) for c in connections]

    # FIXME: Check all the updates and make sure they actually update like they're supposed to.

    def get_application_interface_connection(
        self, application_id: int, interface_id: int, connection_id: int
    ):
        """
        This function returns a single connection for a given interface and application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/242/interfaces/31/connections/494

        Usage:
            connection = sa.get_application_interface_connection(242, 31, 494)
        """

        res = self.api.get_application_interface_connection(
            application_id, interface_id, connection_id
        )
        data = self.api.handle_json(res, "get_application_interface_connection")
        data = get_api_node(data, "interface_connection")
        if isinstance(data, list) and data:
            data = data[0]

        return InterfaceConnection.kwargify(data)

    def add_application_interface_connection(
        self,
        application_id: int,
        interface_id: int,
        connection: Optional[InterfaceConnection] = None,
        /,
        *,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        services: Optional[List[Union[Service, ObjectReference, int]]] = None,
        sources: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        destinations: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        open_tickets: Optional[List[Union[ObjectReference, int]]] = None,
        connected_servers: Optional[List[Union[ObjectReference, int]]] = None,
    ):
        """
        This function creates a new connection for a given interface and application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces/31/connections

        Usage:
            Example 1:
            sa.add_application_interface_connection(
                60,
                31,
                name="My Connection",
                comment="This is my connection",
                services=[1, 2],
                sources=[NetworkObject(id=1), NetworkObject(id=2)],
                destinations=[NetworkObject(id=3), NetworkObject(id=4)],
            )

            Example 2:
            sa.add_application_interface_connection(
                60,
                31,
                InterfaceConnection(
                    name="My Connection",
                    comment="This is my connection",
                    services=[Service(id=1), Service(id=2)],
                    sources=[NetworkObject(id=1), NetworkObject(id=2)],
                    destinations=[NetworkObject(id=3), NetworkObject(id=4)],
                )
            )
        """

        services = (
            [s if isinstance(s, int) else s.id for s in services] if services else []
        )
        sources = (
            [s if isinstance(s, int) else s.id for s in sources] if sources else []
        )
        destinations = (
            [d if isinstance(d, int) else d.id for d in destinations]
            if destinations
            else []
        )
        open_tickets = (
            [t if isinstance(t, int) else t.id for t in open_tickets]
            if open_tickets
            else []
        )
        connected_servers = (
            [s if isinstance(s, int) else s.id for s in connected_servers]
            if connected_servers
            else []
        )

        if not connection:
            connection = InterfaceConnection(
                name=name,
                comment=comment,
                services=[ObjectReference(id=s) for s in services],
                sources=[ObjectReference(id=s) for s in sources],
                destinations=[ObjectReference(id=d) for d in destinations],
                open_tickets=[ObjectReference(id=t) for t in open_tickets],
                connected_servers=[ObjectReference(id=s) for s in connected_servers],
            )

        res = self.api.add_application_interface_connection(
            application_id, interface_id, connection._json
        )
        return self.api.handle_creation(
            res, "add_application_interface_connection", cls=InterfaceConnection
        )

    def update_application_interface_connection(
        self,
        application_id: int,
        interface_id: int,
        connection_id: int | InterfaceConnection,
        /,
        *,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        services: Optional[List[Union[Service, ObjectReference, int]]] = None,
        sources: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        destinations: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        open_tickets: Optional[List[Union[ObjectReference, int]]] = None,
        connected_servers: Optional[List[Union[ObjectReference, int]]] = None,
        force_empty_lists: bool = False,
    ):
        """
        This function updates an existing connection for a given interface and application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces/31/connections/494

        Usage:
            Example 1:
            sa.update_application_interface_connection(
                60,
                31,
                494,
                name="My Connection",
                comment="This is my connection",
                services=[1, 2],
                sources=[NetworkObject(id=1), NetworkObject(id=2)],
                destinations=[NetworkObject(id=3), NetworkObject(id=4)],
            )

            Example 2:
            sa.update_application_interface_connection(
                60,
                31,
                InterfaceConnection(
                    id=494,
                    name="My Connection",
                    comment="This is my connection",
                    services=[Service(id=1), Service(id=2)],
                    sources=[NetworkObject(id=1), NetworkObject(id=2)],
                    destinations=[NetworkObject(id=3), NetworkObject(id=4)],
                )
            )
        """

        connection = None
        if isinstance(connection_id, InterfaceConnection):
            connection = connection_id
            connection_id = connection_id.id
        else:
            services = (
                [s if isinstance(s, int) else s.id for s in services]
                if services
                else []
            )
            sources = (
                [s if isinstance(s, int) else s.id for s in sources] if sources else []
            )
            destinations = (
                [d if isinstance(d, int) else d.id for d in destinations]
                if destinations
                else []
            )
            open_tickets = (
                [t if isinstance(t, int) else t.id for t in open_tickets]
                if open_tickets
                else []
            )
            connected_servers = (
                [s if isinstance(s, int) else s.id for s in connected_servers]
                if connected_servers
                else []
            )

            if not force_empty_lists and (
                not services or not sources or not destinations
            ):
                raise ValueError(
                    "services, sources, and destinations should not be empty. They will be overwritten. If this was intentional, set force_empty_lists=True"
                )

            connection = InterfaceConnection(
                name=name,
                comment=comment,
                services=[ObjectReference(id=s) for s in services],
                sources=[ObjectReference(id=s) for s in sources],
                destinations=[ObjectReference(id=d) for d in destinations],
                open_tickets=[ObjectReference(id=t) for t in open_tickets],
                connected_servers=[ObjectReference(id=s) for s in connected_servers],
            )

        res = self.api.update_application_interface_connection(
            application_id, interface_id, connection_id, connection._json
        )
        self.api.handle_response(
            res, "update_application_interface_connection", action="update"
        )

    def delete_application_interface_connection(
        self, application_id: int, interface_id: int, connection_id: int
    ):
        """
        This function deletes a connection for a given interface and application.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/interfaces/31/connections/494

        Usage:
            sa.delete_application_interface_connection(60, 31, 494)
        """

        res = self.api.delete_application_interface_connection(
            application_id, interface_id, connection_id
        )
        self.api.handle_response(
            res, "delete_application_interface_connection", action="delete"
        )

    def get_application_connections_to_applications(self, application_id: int):
        """
        This function returns a list of all connections to other applications for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_to_applications

        Usage:
            connections = sa.get_application_connections_to_applications(60)
        """

        res = self.api.get_application_connections_to_applications(application_id)
        data = self.api.handle_json(res, "get_application_connections_to_applications")
        connections = get_api_node(
            data, "connections_to_applications.connection_to_application", listify=True
        )
        return [ApplicationInterfaceInstanceGet.kwargify(c) for c in connections]

    def get_application_connection_to_application(
        self, application_id: int, connection_id: int
    ):
        """
        This function returns a single connection to another application for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_to_applications/100

        Usage:
            connection = sa.get_application_connection_to_application(60, 100)
        """

        res = self.api.get_application_connection_to_application(
            application_id, connection_id
        )
        data = self.api.handle_json(res, "get_application_connection_to_application")
        data = get_api_node(data, "connection_to_application")
        if isinstance(data, list) and data:
            data = data[0]

        return ApplicationInterfaceInstanceGet.kwargify(data)

    def add_application_connection_to_application(
        self,
        application_id: int,
        conn: ApplicationInterfaceInstance | None = None,
        /,
        *,
        comment: Optional[str] = None,
        name: Optional[str] = None,
        servers: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        application_interface_id: Optional[int] = None,
    ):
        """
        This function creates a new connection to another application for a given application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_to_applications

        Usage:
            Example 1:
            sa.add_application_connection_to_application(
                60,
                name="My Connection",
                comment="This is my connection",
                servers=[1],
                application_interface_id=100,
            )

            Example 2:
            sa.add_application_connection_to_application(
                60,
                ApplicationInterfaceInstance(
                    name="My Connection",
                    comment="This is my connection",
                    servers=[ObjectReference(id=1)],
                    application_interface_id=100,
                )
            )
        """

        servers = (
            [s if isinstance(s, int) else s.id for s in servers] if servers else []
        )

        if not conn:
            conn = ApplicationInterfaceInstance(
                comment=comment,
                name=name,
                servers=[ObjectReference(id=s) for s in servers],
                application_interface_id=application_interface_id,
            )

        if not conn.name:
            raise ValueError("name is required")
        if not conn.application_interface_id:
            raise ValueError("application_interface_id is required")

        res = self.api.add_application_connection_to_application(
            application_id, conn._json
        )
        return self.api.handle_creation(
            res,
            "add_application_connection_to_application",
            cls=ApplicationInterfaceInstanceGet,
        )

    def update_application_connection_to_application(
        self,
        application_id: int,
        connection_id: int | ApplicationInterfaceInstance,
        /,
        *,
        comment: Optional[str] = None,
        name: Optional[str] = None,
        servers: Optional[List[Union[NetworkObject, ObjectReference, int]]] = None,
        application_interface_id: Optional[int] = None,
        force_empty_servers: bool = False,
    ):
        """
        This function updates an existing connection to another application for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_to_applications/100

        Usage:
            Example 1:
            sa.update_application_connection_to_application(
                60,
                100,
                name="My Connection",
                comment="This is my connection",
                connections=[Connection(services=[ObjectReference(id=1)])],
            )

            Example 2:
            sa.update_application_connection_to_application(
                60,
                ApplicationInterfaceInstance(
                    id=100,
                    name="My Connection",
                    comment="This is my connection",
                    servers=[ObjectReference(id=1)],
                    application_interface_id=100,
                )
            )
        """

        if isinstance(connection_id, ApplicationInterfaceInstanceGet):
            raise ValueError(
                "The SDK does not support updating from an ApplicationInterfaceInstanceGet object fetched from the API. Please use an ApplicationInterfaceInstance object instead."
            )

        if isinstance(connection_id, ApplicationInterfaceInstance):
            connection = connection_id
            connection_id = connection_id.id
        else:
            servers = (
                [s if isinstance(s, int) else s.id for s in servers] if servers else []
            )

            connection = ApplicationInterfaceInstance(
                comment=comment,
                name=name,
                servers=[ObjectReference(id=s) for s in servers],
                application_interface_id=application_interface_id,
            )

            if not connection.servers and not force_empty_servers:
                raise ValueError(
                    "servers is empty. If this is intentional, set force_empty_servers=True"
                )

        if connection.name is None:
            raise ValueError("name is required")

        res = self.api.update_application_connection_to_application(
            application_id, connection_id, connection._json
        )

        self.api.handle_response(
            res, "update_application_connection_to_application", action="update"
        )

    def delete_application_connection_to_application(
        self, application_id: int, connection_id: int
    ):
        """
        This function deletes a connection to another application for a given application.

        Method: DELETE
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/connections_to_applications/100

        Usage:
            sa.delete_application_connection_to_application(60, 100)
        """

        res = self.api.delete_application_connection_to_application(
            application_id, connection_id
        )
        self.api.handle_response(
            res, "delete_application_connection_to_application", action="delete"
        )

    def get_application_history(
        self,
        application_id: int,
        start_date: DateFilterType = None,
        end_date: DateFilterType = None,
        count: Optional[int] = None,
        start: Optional[int] = None,
        type: Optional[str] = None,
        user: Optional[str] = None,
        show_members: Optional[bool] = None,
    ):
        """
        This function returns the history of a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/history

        Usage:
            history = sa.get_application_history(60)
        """

        start_date = coerce_date_filter_type(start_date)
        end_date = coerce_date_filter_type(end_date)

        res = self.api.get_application_history(
            application_id,
            start_date=start_date,
            end_date=end_date,
            count=count,
            start=start,
            type=type,
            user=user,
            show_members=show_members,
        )
        data = self.api.handle_json(res, "get_application_history")
        history = get_api_node(data, "history_records.history_record", listify=True)
        return [HistoryRecord.kwargify(h) for h in history]

    def get_application_access_requests(self, application_id: int):
        """
        This function returns a list of all access requests for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/application_access_requests

        Usage:
            access_requests = sa.get_application_access_requests(60)
        """

        res = self.api.get_application_access_requests(application_id)
        data = self.api.handle_json(res, "get_application_access_requests")
        return [
            ApplicationAccessRequest.kwargify(a)
            for a in get_api_node(
                data,
                "application_access_requests.application_access_request",
                listify=True,
            )
        ]

    def get_application_access_request(self, application_id: int, request_id: int):
        """
        This function returns a single access request for a given application.

        Method: GET
        URL: /securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests/1

        Usage:
            access_request = sa.get_application_access_request(8, 1)
        """

        res = self.api.get_application_access_request(application_id, request_id)
        data = self.api.handle_json(res, "get_application_access_request")
        data = get_api_node(data, "application_access_request")
        return ApplicationAccessRequest.kwargify(
            data[0] if isinstance(data, list) else data
        )

    def add_application_access_request(
        self,
        application_id: int,
        request: Optional[ApplicationAccessRequest] = None,
        /,
        *,
        comment: Optional[str] = None,
        action: Optional[str] = None,
        server_ip: Optional[str | IPAddress] = None,
        server_group_id: Optional[str | int] = None,
        server_group_name: Optional[str] = None,
    ):
        """
        This function creates a new access request for a given application.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/application_access_requests

        Usage:
            Example 1:
            sa.add_application_access_request(
                8,
                comment="Please grant me access to this application",
                server_ip='1.1.2.2',
                server_group_id='1',
            )

            Example 2:
            sa.add_application_access_request(
                8,
                request=ApplicationAccessRequest(
                    comment="Please grant me access to this application",
                    server_ip='1.1.2.2',
                    server_group_id='1',
                ),
            )

        """

        request = (
            ApplicationAccessRequest(
                comment=comment,
                action=action,
                server_ip=server_ip,
                server_group_id=server_group_id,
                server_group_name=server_group_name,
            )
            if request is None
            else request
        )

        res = self.api.add_application_access_request(application_id, request._json)
        return self.api.handle_creation(
            res, "add_application_access_request", cls=ApplicationAccessRequest
        )

    def approve_application_access_request(
        self, application_id: int, request_id: int | ApplicationAccessRequest
    ):
        """
        This function approves an access request for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/application_access_requests/1

        Usage:
            Example 1:
            sa.approve_application_access_request(60, 1)

            Example 2:
            sa.approve_application_access_request(60, ApplicationAccessRequest(id=1))
        """

        if isinstance(request_id, ApplicationAccessRequest):
            request_id = request_id.id
        if request_id is None:
            raise ValueError("request_id is required")

        res = self.api.update_application_access_request(
            application_id, request_id, {"action": "APPROVE"}
        )

        self.api.handle_response(
            res, "approve_application_access_request", action="update"
        )

    def reject_application_access_request(
        self, application_id: int, request_id: int | ApplicationAccessRequest
    ):
        """
        This function rejects an access request for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/60/application_access_requests/1

        Usage:
            Example 1:
            sa.reject_application_access_request(60, 1)

            Example 2:
            sa.reject_application_access_request(60, ApplicationAccessRequest(id=1))
        """

        if isinstance(request_id, ApplicationAccessRequest):
            request_id = request_id.id
        if request_id is None:
            raise ValueError("request_id is required")

        res = self.api.update_application_access_request(
            application_id, request_id, {"action": "REJECT"}
        )

        self.api.handle_response(
            res, "reject_application_access_request", action="update"
        )

    def bulk_update_application_access_requests(
        self, application_id: int, requests: List[ApplicationAccessRequest]
    ):
        """
        This function updates multiple access requests for a given application.

        Method: PUT
        URL: /securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests/1

        Usage:
            sa.bulk_update_application_access_requests(
                60,
                [
                    ApplicationAccessRequest(id=1, action="APPROVE"),
                    ApplicationAccessRequest(id=2, action="REJECT"),
                ]
            )
        """

        for r in requests:
            if r.action not in [
                ApplicationAccessRequest.Action.APPROVE,
                ApplicationAccessRequest.Action.REJECT,
            ]:
                raise ValueError(
                    f"Action for application_access_requests must be either 'APPROVE' or 'REJECT'. Got: {r.action}"
                )

        res = self.api.update_application_access_requests(
            application_id, [r._json for r in requests]
        )
        self.api.handle_response(
            res, "update_application_access_requests", action="update"
        )

    def repair_connection(
        self,
        application_id: int,
        connection_id: int,
        workflow: str,
        subject: str,
        fields: Optional[List[Field]] = None,
    ) -> Ticket:
        """
        This function repairs a connection.

        It takes a workflow name, a list of fields, and a subject and produces a new ticket.

        You may need to pass in metadata fields, like TextField or TextArea.

        Method: POST
        URL: /securechangeworkflow/api/secureapp/repository/connections/1/repair

        Usage:
            sa.repair_connection(1, 1, workflow="Firewall Change Request", subject="Repair Connection", fields=[Field(name="Field1", value="Value1")])
        """

        ticket = Ticket.create(workflow, subject)
        step = ticket.create_step(None)
        task = step.create_task()
        task.fields.extend(fields or [])

        res = self.api.repair_connection(
            application_id, connection_id, ticket.post_json
        )
        return self.api.handle_creation(res, "repair_connection", cls=Ticket)



from datetime import datetime
from netaddr import IPAddress
import pytest
import json
from pytos2.api import ApiError
from pytos2.secureapp.access_requests import ApplicationAccessRequest
from pytos2.secureapp.applications import ApplicationInterfaceInstance
import responses

from pytos2.secureapp.history import (
    HistoryConnectionBaseSnapshot,
    HistoryConnectionDetails,
)
from pytos2.securetrack.service_object import TCPServiceObject

from . import conftest  # noqa

from pytos2.secureapp.entrypoint import Sa
from pytos2.secureapp.application_identities import ApplicationIdentity
from dateutil.tz import tzoffset


class TestEntrypoint:
    @responses.activate
    def test_application_identities(self, sa: Sa, application_identities_mock):
        # print(dir(sa))
        identity = sa.application_identities[0]
        assert isinstance(identity, ApplicationIdentity)

    @responses.activate
    def test_application_identites_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.application_identities

    @responses.activate
    def test_applications(self, sa: Sa, applications_mock):
        applications = sa.get_applications()
        assert applications[0].id == 60
        assert (
            applications[0].connections[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126"
        )
        assert applications[0].owner and applications[0].owner.name == "r"

    @responses.activate
    def test_application(self, sa: Sa, applications_mock):
        application = sa.get_application(54)
        assert application.id == 54
        assert application.owner.name == "Jessica.Sanchez"
        assert (
            application.open_tickets[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/35"
        )

    @responses.activate
    def test_update_application(self, sa: Sa, applications_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54",
            status=200,
            body="",
        )

        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "There is no application with the specified ID 1337.",
                }
            },
        )

        application = sa.get_application(54)
        application.name = "New Name"

        sa.update_application(application)
        with pytest.raises(ApiError):
            sa.update_application(1337, name="New Name")

        sa.update_application(54, name="New Name", owner=7)

    @responses.activate
    def test_create_application(self, scw, sa: Sa, applications_mock, users_mock):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54"
            },
        )

        application = sa.add_application(
            name="VPN users access to RnD users", owner="r"
        )

        # Input and output don't really matter because we're mocking the response
        assert application.id == 54
        assert application.owner.name == "Jessica.Sanchez"
        assert application.name == "VPN users access to RnD users"

    @responses.activate
    def test_delete_application(self, sa: Sa, applications_mock):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/54",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "There is no application with the specified ID 1337.",
                }
            },
        )

        sa.delete_application(54)

        with pytest.raises(ApiError):
            sa.delete_application(1337)

    @responses.activate
    def test_bulk_update_applications(self, sa: Sa, applications_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/",
            status=200,
            body="",
        )

        app = sa.get_application(54)

        sa.bulk_update_applications([app])

    @responses.activate
    def test_application_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application(54)

        with pytest.raises(ValueError):
            sa.get_applications()

    @responses.activate
    def test_application_connections(self, sa: Sa, application_connections_mock):
        connections = sa.get_application_connections(60)
        assert connections[0].id == 126
        assert connections[0].name == "Connection 1"
        assert connections[0].comment == ""
        assert connections[0].external is False
        assert connections[0].sources[0].id == 762
        assert connections[0].sources[0].name == "CRM_01"
        assert connections[0].sources[0].display_name == "CRM_01"
        assert connections[0].sources[0].link.href == (
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/network_objects/762"
        )

        assert connections[0].destinations[0].id == 998
        assert connections[0].destinations[0].name == "Finance_192.168.50.100"
        assert connections[0].destinations[0].display_name == "Finance_192.168.50.100"
        assert connections[0].destinations[0].type == "host"

    @responses.activate
    def test_application_connections_extended(
        self, sa: Sa, application_connections_mock
    ):
        connections = sa.get_extended_application_connections(60)
        assert connections[0].id == 237
        assert connections[0].name == "Connection 4"
        assert connections[0].comment == ""
        assert connections[0].external is False
        assert connections[0].sources[0].id == 762
        assert connections[0].sources[0].name == "CRM_01"
        assert connections[0].sources[0].display_name == "CRM_01"
        assert connections[0].sources[0].xsi_type.value == "hostNetworkObjectDTO"

    @responses.activate
    def test_application_connection(self, sa: Sa, application_connections_mock):
        connection = sa.get_application_connection(60, 126)
        assert connection.name == "Connection 1"
        assert connection.sources[0].name == "CRM_01"

    @responses.activate
    def test_application_connection_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application_connection(60, 126)

        with pytest.raises(ValueError):
            sa.get_application_connections(60)

    @responses.activate
    def test_add_application_connection(self, sa: Sa, application_connections_mock):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126"
            },
        )

        connection = sa.add_application_connection(
            60, name="Connection 1", sources=[762]
        )

        assert connection.id == 126
        assert connection.name == "Connection 1"
        assert connection.sources[0].name == "CRM_01"
        assert connection.sources[0].id == 762
        assert connection.services[0].name == "ftp"
        assert connection.services[0].id == 44

    @responses.activate
    def test_update_application_connection(self, sa: Sa, application_connections_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126",
            status=200,
            body="",
        )

        conn = sa.get_application_connection(60, 126)
        sa.update_application_connection(60, conn)
        sa.update_application_connection(60, 126, name="New Name")

    @responses.activate
    def test_bulk_update_application_connections(
        self, sa: Sa, application_connections_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections",
            status=200,
            body="",
        )

        conn = sa.get_application_connection(60, 126)
        sa.bulk_update_application_connections(60, [conn])

    @responses.activate
    def test_delete_application_connection(self, sa: Sa, application_connections_mock):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "Connection with ID 1337 was not found for application with ID 60.",
                }
            },
        )

        sa.delete_application_connection(60, 126)

        with pytest.raises(ApiError):
            sa.delete_application_connection(60, 1337)

    @responses.activate
    def test_application_interfaces(self, sa: Sa, application_interfaces_mock):
        interfaces = sa.get_application_interfaces(242)
        assert interfaces[0].id == 31
        assert interfaces[0].name == "Interface to DNS clients"
        assert interfaces[0].comment == ""
        assert interfaces[0].is_published is False
        assert interfaces[0].application_id == 242
        assert interfaces[0].interface_connections[0].name == "DNS clients"
        assert interfaces[0].interface_connections[0].id == 494
        assert interfaces[0].interface_connections[0].open_tickets == []
        assert interfaces[0].interface_connections[0].services[0].name == "domain-udp"
        assert (
            interfaces[0].interface_connections[0].destinations[0].name
            == "DNS Server 1"
        )
        assert (
            interfaces[0].interface_connections[0].connected_servers[0].name
            == "Mail servers"
        )

    @responses.activate
    def test_get_application_interface(self, sa: Sa, application_interfaces_mock):
        # The reason these are different from above is that the data is from different
        # points in time, between which points I mutated data on the server from which
        # this data is.

        interface = sa.get_application_interface(242, 31)
        assert interface.id == 31
        assert interface.name == "Interface to DNS clients"
        assert interface.comment == ""
        assert interface.is_published is False
        assert interface.application_id == 242
        assert interface.interface_connections[0].name == "My Connection3"
        assert interface.interface_connections[0].id == 20117
        assert interface.interface_connections[0].open_tickets == []
        assert (
            interface.interface_connections[0].connected_servers[0].name
            == "Mail servers"
        )

    @responses.activate
    def test_add_application_interface(self, sa: Sa, application_interfaces_mock):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31"
            },
        )

        interface = sa.add_application_interface(
            242, name="Interface to DNS clients", comment=""
        )
        assert interface.id == 31
        assert interface.name == "Interface to DNS clients"
        assert interface.comment == ""

    @responses.activate
    def test_update_application_interface(self, sa: Sa, application_interfaces_mock):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31",
            status=200,
            body="",
        )

        interface = sa.get_application_interface(242, 31)
        sa.update_application_interface(242, interface)

        sa.update_application_interface(242, 31, name="New Name")

    @responses.activate
    def test_delete_application_interface(self, sa: Sa, application_interfaces_mock):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "Application interface with ID 1337 was not found for application with ID 242.",
                }
            },
        )

        interface = sa.get_application_interface(242, 31)
        sa.delete_application_interface(242, 31)

        with pytest.raises(ApiError):
            sa.delete_application_interface(242, 1337)

    @responses.activate
    def test_get_application_interface_connections(
        self, sa: Sa, application_interface_connections_mock
    ):
        connections = sa.get_application_interface_connections(242, 31)
        assert connections[0].id == 20116
        assert connections[0].name == "My Connection3"
        assert connections[0].sources[0].name == "DNS Server 1"
        assert connections[0].connected_servers[0].name == "Mail servers"
        assert connections[0].services[0].name == "domain-udp"

    @responses.activate
    def test_get_application_interface_connection(
        self, sa: Sa, application_interface_connections_mock
    ):
        connection = sa.get_application_interface_connection(242, 31, 494)
        assert connection.id == 494
        assert connection.name == "DNS clients"
        assert connection.destinations[0].name == "DNS Server 1"
        assert connection.services[0].name == "domain-udp"

    @responses.activate
    def test_add_application_interface_connection(
        self, sa: Sa, application_interface_connections_mock
    ):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31/interface_connections",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31/interface_connections/494"
            },
        )

        connection = sa.add_application_interface_connection(
            242,
            31,
            name="DNS clients",
            destinations=[4145, 4181, 5812],
            connected_servers=[4173],
            services=[28],
        )
        assert connection.id == 494
        assert connection.name == "DNS clients"

    @responses.activate
    def test_update_application_interface_connection(
        self, sa: Sa, application_interface_connections_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31/interface_connections/494",
            status=200,
            body="",
        )

        connection = sa.get_application_interface_connection(242, 31, 494)
        sa.update_application_interface_connection(242, 31, connection)

        with pytest.raises(ValueError):
            sa.update_application_interface_connection(
                242, 31, 494, name="New Name", destinations=[4145, 4181, 5812]
            )
        sa.update_application_interface_connection(
            242,
            31,
            494,
            name="New Name",
            destinations=[4145, 4181, 5812],
            connected_servers=[4173],
            services=[28],
            force_empty_lists=True,
        )

    @responses.activate
    def test_delete_application_interface_connection(
        self, sa: Sa, application_interface_connections_mock
    ):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31/interface_connections/494",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/242/application_interfaces/31/interface_connections/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "Application interface connection with ID 1337 was not found for application interface with ID 31.",
                }
            },
        )

        sa.delete_application_interface_connection(242, 31, 494)

        with pytest.raises(ApiError):
            sa.delete_application_interface_connection(242, 31, 1337)

    @responses.activate
    def test_get_application_connections_to_applications(
        self, sa: Sa, application_connections_to_applications_mock
    ):
        connections = sa.get_application_connections_to_applications(8)
        assert connections[0].id == 2
        assert connections[0].name == "AD access 01"
        assert connections[0].comment == ""
        assert connections[0].application_id == 8
        assert connections[0].connections[0].external is False
        assert connections[0].connections[0].status == "DISCONNECTED"
        assert connections[0].connections[0].id == 65
        assert connections[0].connections[0].name == "Access From AD"
        assert connections[0].connections[0].open_tickets[0].id == 21
        assert (
            connections[0].connections[0].open_tickets[0].name == "Active Directory #2"
        )

    @responses.activate
    def test_get_application_connection_to_application(
        self, sa: Sa, application_connections_to_applications_mock
    ):
        connection = sa.get_application_connection_to_application(8, 3)
        assert connection.id == 3
        assert connection.name == "AD access 02"
        assert connection.comment == ""
        assert connection.application_id == 8
        assert connection.connections[0].external is False
        assert connection.connections[0].status == "DISCONNECTED"
        assert connection.connections[0].id == 66
        assert connection.connections[0].name == "Access To AD"
        assert connection.connections[0].open_tickets[0].id == 21
        assert connection.connections[0].open_tickets[0].name == "Active Directory #2"

    @responses.activate
    def test_add_application_connection_to_application(
        self, sa: Sa, application_connections_to_applications_mock
    ):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/connections_to_applications",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/connections_to_applications/3"
            },
        )

        connection = sa.add_application_connection_to_application(
            8, name="AD access 02", application_interface_id=6, servers=[8]
        )
        assert connection.id == 3

        connection = sa.add_application_connection_to_application(
            8,
            ApplicationInterfaceInstance(
                name="AD access 02", application_interface_id=6, servers=[8]
            ),
        )

        assert connection.id == 3

        with pytest.raises(ValueError):
            sa.add_application_connection_to_application(8)
        with pytest.raises(ValueError):
            sa.add_application_connection_to_application(8, name="AD access 02")

    @responses.activate
    def test_update_application_connection_to_application(
        self, sa: Sa, application_connections_to_applications_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/connections_to_applications/3",
            status=200,
            body="",
        )

        with pytest.raises(ValueError):
            # Missing name
            sa.update_application_connection_to_application(
                8, 3, application_interface_id=6, servers=[8]
            )

        connection = sa.get_application_connection_to_application(8, 3)
        with pytest.raises(ValueError):
            sa.update_application_connection_to_application(8, connection)

        sa.update_application_connection_to_application(
            8,
            ApplicationInterfaceInstance(
                id=3, name="New Name", application_interface_id=6
            ),
        )

        with pytest.raises(ValueError):
            sa.update_application_connection_to_application(
                8, 3, name="New Name", application_interface_id=6
            )

        sa.update_application_connection_to_application(
            8, 3, name="New Name", application_interface_id=6, servers=[8]
        )

    @responses.activate
    def test_delete_application_connection_to_application(
        self, sa: Sa, application_connections_to_applications_mock
    ):
        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/connections_to_applications/3",
            status=200,
            body="",
        )

        responses.add(
            responses.DELETE,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/connections_to_applications/1337",
            status=404,
            json={
                "result": {
                    "code": "NOT_FOUND_ERROR",
                    "message": "Connection to application with ID 1337 was not found for application with ID 8.",
                }
            },
        )

        sa.delete_application_connection_to_application(8, 3)

        with pytest.raises(ApiError):
            sa.delete_application_connection_to_application(8, 1337)

    @responses.activate
    def test_application_history(self, sa: Sa, application_history_mock):
        history = sa.get_application_history(60)

        assert history[0].change_description == "Connection deleted"

        assert history[0].change_details.xsi_type.value == "historyConnectionDetailsDTO"
        assert isinstance(history[0].change_details, HistoryConnectionDetails)

        assert history[0].change_details.removed_sources[0].name == "Any"
        assert history[0].change_details.removed_sources[0].id == 1
        assert history[0].change_details.removed_sources[0].display_name == "Any"
        assert (
            history[0].change_details.removed_sources[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/network_objects/1"
        )

        assert history[0].change_details.removed_destinations[0].name == "CRM_01"
        assert history[0].change_details.removed_destinations[0].id == 762
        assert (
            history[0].change_details.removed_destinations[0].display_name == "CRM_01"
        )
        assert (
            history[0].change_details.removed_destinations[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/network_objects/762"
        )

        assert history[0].change_details.removed_services[0].name == "amitay2"
        assert history[0].change_details.removed_services[0].id == 235
        assert history[0].change_details.removed_services[0].display_name == "amitay2"
        assert (
            history[0].change_details.removed_services[0].link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/services/235"
        )

        assert history[0].date == datetime(
            2024, 4, 15, 3, 26, 57, 257000, tzinfo=tzoffset(None, -25200)
        )
        assert history[0].modified_object.display_name == "Connection 2"
        assert history[0].modified_object.id == 203
        assert (
            history[0].modified_object.link.href
            == "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/203"
        )
        assert history[0].modified_object.name == "Connection 2"
        assert history[0].snapshot.comment == ""
        assert history[0].snapshot.id == 203
        assert history[0].snapshot.name == "Connection 2"

        assert isinstance(history[0].snapshot, HistoryConnectionBaseSnapshot)
        assert history[0].snapshot.xsi_type.value == "historyConnectionDTO"

        assert history[0].type == "Connection"
        assert history[0].user.display_name == "Henry Carr"
        assert history[0].user.id == 4
        assert (
            history[0].user.link.href
            == "https://198.18.0.1/securechangeworkflow/api/securechange/users/4"
        )
        assert history[0].user.name == "Henry Carr"

        assert "Connection changed" in history[3].change_description
        assert history[3].snapshot.id == 236
        assert history[3].snapshot.name == "Connection 3"
        assert history[3].snapshot.services[0].name == "http"
        assert isinstance(history[3].snapshot.services[0], TCPServiceObject)
        assert history[3].snapshot.services[0].min_port == 80
        assert history[3].snapshot.services[0].max_port == 80
        assert history[3].snapshot.services[0].protocol == 6

        assert history[3].snapshot.destinations[0].name == "CRM_01"
        assert history[3].snapshot.destinations[0].ip == IPAddress("192.168.205.33")

        # Covering empty case in find_in_detail_list
        assert history[4].change_details.removed_sources == []

        assert history[4].change_details.added_destinations[0].id == 3245
        assert history[4].change_details.added_destinations[0].name == "EDL_server2"
        assert (
            history[4].change_details.added_destinations[0].display_name
            == "EDL_server2"
        )

        assert history[5].change_details.added_services[0].id == 233
        assert history[5].change_details.added_services[0].name == "amitay5"

        assert history[11].change_details.added_sources[0].id == 3244
        assert history[11].change_details.added_sources[0].name == "EDL_subnet"

    @responses.activate
    def test_application_history_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.get_application_history(60)

    @responses.activate
    def test_get_application_access_requests(
        self, sa: Sa, application_access_requests_mock
    ):
        requests = sa.get_application_access_requests(8)

        assert requests[0].id == 1
        assert requests[0].server_group_id == 412
        assert requests[0].server_group_name == "Sales"
        assert requests[0].server_ip == IPAddress("1.1.1.1")
        assert requests[0].comment == ""
        assert requests[0].action.value == "OPENED"

    @responses.activate
    def test_get_application_access_request(
        self, sa: Sa, application_access_requests_mock
    ):
        request = sa.get_application_access_request(8, 1)

        assert request.id == 1
        assert request.server_group_id == 412
        assert request.server_group_name == "Sales"
        assert request.server_ip == IPAddress("1.1.1.1")
        assert request.comment == ""
        assert request.action.value == "OPENED"

    @responses.activate
    def test_add_application_access_request(
        self, sa: Sa, application_access_requests_mock
    ):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests/1"
            },
        )

        request = sa.add_application_access_request(
            8, server_group_id=412, server_ip="1.1.1.1"
        )

        assert request.id == 1
        assert request.server_group_id == 412
        assert request.server_group_name == "Sales"

    @responses.activate
    def test_approve_application_access_request(
        self, sa: Sa, application_access_requests_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests/1",
            status=200,
            body="",
        )

        request = sa.get_application_access_request(8, 1)
        sa.approve_application_access_request(8, 1)
        sa.approve_application_access_request(8, request)

        with pytest.raises(ValueError):
            sa.approve_application_access_request(
                8, ApplicationAccessRequest(action="APPROVE")
            )

    @responses.activate
    def test_reject_application_access_request(
        self, sa: Sa, application_access_requests_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests/1",
            status=200,
            body="",
        )

        request = sa.get_application_access_request(8, 1)
        sa.reject_application_access_request(8, 1)
        sa.reject_application_access_request(8, request)

        with pytest.raises(ValueError):
            sa.reject_application_access_request(
                8, ApplicationAccessRequest(action="REJECT")
            )

    @responses.activate
    def test_bulk_update_application_access_requests(
        self, sa: Sa, application_access_requests_mock
    ):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/8/application_access_requests",
            status=200,
            body="",
        )

        sa.bulk_update_application_access_requests(
            8, [ApplicationAccessRequest(id=1, action="APPROVE")]
        )

        with pytest.raises(ValueError):
            sa.bulk_update_application_access_requests(
                8, [ApplicationAccessRequest(id=1, action="OPENED")]
            )

    @responses.activate
    def test_repair_connection(self, sa: Sa, application_connections_mock):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/secureapp/repository/applications/60/connections/126/repair",
            status=201,
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1067"
            },
        )

        with open("tests/securechange/json/ticket/ticket-1067.json") as f:
            ticket_json = json.load(f)

        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1067",
            json=ticket_json,
        )

        ticket = sa.repair_connection(
            60, 126, "Firewall Change Request Workflow", "Repairing connection"
        )
        assert ticket.id == 1067
        assert ticket.workflow.name == "Firewall Change Request Workflow"

        assert ticket.last_task.access_request.name == "Required Access"



from typing import Optional, List

from netaddr import IPAddress, IPNetwork, IPRange  # type: ignore

from pytos2.utils import propify, prop, stringify_optional_obj
from pytos2.models import Jsonable


@propify
class TopologySyncStatus(Jsonable):
    percentage: Optional[int] = prop(None)
    description: Optional[str] = prop(None)


@propify
class TopologyMode(Jsonable):
    domain_id: int = prop(None, key="domainId")
    device_id: int = prop(None, key="mgmtId")
    mode: str = prop("")  # Will be "DISABLED" or "ENABLED"


@propify
class TopologySubnet(Jsonable):
    name: str = prop("")
    ip: IPAddress = prop(None, converter=IPAddress, jsonify=stringify_optional_obj)
    mask: IPAddress = prop(None, converter=IPAddress, jsonify=stringify_optional_obj)
    domain_id_tag: Optional[int] = prop(None, key="domainIdTag")


@propify
class SubnetDevice(Jsonable):
    is_generic: bool = prop(None)
    name: str = prop("")
    interface_id: Optional[int] = prop(None)
    interface_name: str = prop("")
    interface_ip: Optional[IPAddress] = prop(
        None, converter=IPAddress, jsonify=stringify_optional_obj
    )


@propify
class TopologySubnetDetailed(TopologySubnet):
    attached_devices: List[SubnetDevice] = prop(factory=list, flatify="attached_device")
    join_candidates: List[SubnetDevice] = prop(factory=list, flatify="join_candidate")



from enum import Enum
from typing import Optional

from netaddr import IPAddress
from attr.converters import optional
from pytos2.utils import prop, propify, stringify_optional_obj
from pytos2.models import Jsonable


@propify
class ApplicationAccessRequest(Jsonable):
    class Meta(Enum):
        ROOT = "application_access_request"

    class Action(str, Enum):
        APPROVE = "APPROVE"
        REJECT = "REJECT"
        OPENED = "OPENED"
        DENIED = "DENIED"
        WAITING_TICKET = "WAITING_TICKET"
        IN_TICKET = "IN_TICKET"
        CONFIRMED = "CONFIRMED"
        IMMEDIATE_CONFIRM = "IMMEDIATE_CONFIRM"
        IMMEDIATE_DENIED = "IMMEDIATE_DENIED"

    comment: Optional[str] = prop(None)

    action: Optional[Action] = prop(None)
    server_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    server_group_id: Optional[int | str] = prop(None)
    server_group_name: Optional[str] = prop(None)



from enum import Enum
from typing import Optional

from requests import HTTPError, JSONDecodeError

from pytos2.api import BaseAPI, get_app_api_session
from pytos2.secureapp.access_requests import ApplicationAccessRequest
from pytos2.utils import setup_logger


LOGGER = setup_logger("sa_api")


class SaAPI(BaseAPI):
    class Meta(Enum):
        PATH = "securechangeworkflow/api/secureapp"
        APP = "SCW"
        TOS2_ENV = "SC_SERVER_SERVICE"

    def __init__(
        self, hostname: Optional[str], username: Optional[str], password: Optional[str]
    ):
        self.hostname, self.username, self.password, session = get_app_api_session(
            app=self, hostname=hostname, username=username, password=password
        )

        super().__init__(session)

    def get_application_identities(self):
        return self.session.get("repository/application_identities")

    def get_applications(self):
        return self.session.get("repository/applications")

    def get_application(self, application_id):
        return self.session.get(f"repository/applications/{application_id}")

    def add_application(self, application):
        apps = [application] if not isinstance(application, list) else application

        r = self.session.post(
            "repository/applications", json={"applications": {"application": apps}}
        )
        return r

    def update_application(self, application_id, application):
        return self.session.put(
            f"repository/applications/{application_id}",
            json={"application": application},
        )

    def update_applications(self, applications):
        return self.session.put(
            "repository/applications/",
            json={"applications": {"application": applications}},
        )

    def delete_application(self, application_id):
        return self.session.delete(f"repository/applications/{application_id}")

    def get_application_connections(self, application_id):
        return self.session.get(f"repository/applications/{application_id}/connections")

    def get_application_connection(self, application_id, connection_id):
        r = self.session.get(
            f"repository/applications/{application_id}/connections/{connection_id}",
        )
        return r

    def get_extended_application_connections(self, application_id):
        return self.session.get(
            f"repository/applications/{application_id}/connections_extended"
        )

    def add_application_connection(self, application_id, connection):
        return self.session.post(
            f"repository/applications/{application_id}/connections",
            json={"connections": {"connection": [connection]}},
        )

    def update_application_connection(self, application_id, connection_id, connection):
        return self.session.put(
            f"repository/applications/{application_id}/connections/{connection_id}",
            json={"connection": connection},
        )

    def update_application_connections(self, application_id, connections):
        return self.session.put(
            f"repository/applications/{application_id}/connections",
            json={"connections": {"connection": connections}},
        )

    def delete_application_connection(self, application_id, connection_id):
        return self.session.delete(
            f"repository/applications/{application_id}/connections/{connection_id}"
        )

    def get_application_interfaces(self, application_id):
        return self.session.get(
            f"repository/applications/{application_id}/application_interfaces"
        )

    def get_application_interface(self, application_id, interface_id):
        return self.session.get(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}"
        )

    def add_application_interface(self, application_id, interface):
        return self.session.post(
            f"repository/applications/{application_id}/application_interfaces",
            json={"application_interface": interface},
        )

    def update_application_interface(self, application_id, interface_id, interface):
        return self.session.put(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}",
            json={"application_interface": interface},
        )

    def delete_application_interface(self, application_id, interface_id):
        return self.session.delete(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}"
        )

    def get_application_interface_connections(self, application_id, interface_id):
        return self.session.get(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}/interface_connections"
        )

    def get_application_interface_connection(
        self, application_id, interface_id, connection_id
    ):
        return self.session.get(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}/interface_connections/{connection_id}"
        )

    def add_application_interface_connection(
        self, application_id, interface_id, connection
    ):
        return self.session.post(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}/interface_connections",
            json={"interface_connections": {"interface_connection": [connection]}},
        )

    def update_application_interface_connection(
        self, application_id, interface_id, connection_id, connection
    ):
        return self.session.put(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}/interface_connections/{connection_id}",
            json={"interface_connection": connection},
        )

    def delete_application_interface_connection(
        self, application_id, interface_id, connection_id
    ):
        return self.session.delete(
            f"repository/applications/{application_id}/application_interfaces/{interface_id}/interface_connections/{connection_id}"
        )

    def get_application_connections_to_applications(self, application_id):
        return self.session.get(
            f"repository/applications/{application_id}/connections_to_applications"
        )

    def get_application_connection_to_application(self, application_id, connection_id):
        return self.session.get(
            f"repository/applications/{application_id}/connections_to_applications/{connection_id}"
        )

    def add_application_connection_to_application(self, application_id, connection):
        return self.session.post(
            f"repository/applications/{application_id}/connections_to_applications",
            json={"connection_to_application": connection},
        )

    def update_application_connection_to_application(
        self, application_id, connection_id, connection
    ):
        return self.session.put(
            f"repository/applications/{application_id}/connections_to_applications/{connection_id}",
            json={"connection_to_application": connection},
        )

    def delete_application_connection_to_application(
        self, application_id, connection_id
    ):
        return self.session.delete(
            f"repository/applications/{application_id}/connections_to_applications/{connection_id}"
        )

    def get_application_history(
        self,
        application_id,
        start_date=None,
        end_date=None,
        count=None,
        start=None,
        type=None,
        user=None,
        show_members=None,
    ):
        return self.session.get(
            f"repository/applications/{application_id}/history",
            params={
                "start_date": start_date,
                "end_date": end_date,
                "count": count,
                "start": start,
                "type": type,
                "user": user,
                "showMembers": show_members,
            },
        )

    def get_application_access_requests(self, application_id):
        return self.session.get(
            f"repository/applications/{application_id}/application_access_requests"
        )

    def get_application_access_request(self, application_id, request_id):
        return self.session.get(
            f"repository/applications/{application_id}/application_access_requests/{request_id}"
        )

    def add_application_access_request(self, application_id, request: dict):
        return self.session.post(
            f"repository/applications/{application_id}/application_access_requests",
            json={"application_access_request": request},
        )

    def update_application_access_request(
        self, application_id, request_id, request: dict
    ):
        return self.session.put(
            f"repository/applications/{application_id}/application_access_requests/{request_id}",
            json={"application_access_request": request},
        )

    def update_application_access_requests(self, application_id, requests: list):
        return self.session.put(
            f"repository/applications/{application_id}/application_access_requests",
            json={
                "application_access_requests": {"application_access_request": requests}
            },
        )

    def delete_application_access_request(self, application_id, request_id):
        return self.session.delete(
            f"repository/applications/{application_id}/application_access_requests/{request_id}"
        )

    def repair_connection(self, application_id, connection_id, ticket):
        return self.session.post(
            f"repository/applications/{application_id}/connections/{connection_id}/repair",
            json=ticket,
        )



from typing import Any, Optional, Tuple
from .models import GraphResponse, Variables

from ..api import GraphqlError, OAuth2APISession


class BaseGql:
    hostname: Optional[str]
    username: Optional[str]
    password: Optional[str]
    session: Optional[OAuth2APISession]

    def query(
        self, query: str, variables: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        return self.requestgql(query, variables=variables)

    def mutate(
        self, query: str, variables: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        raise NotImplementedError("mutate not supported in pytos2-ce")

    def requestgql(self, query: str, variables: Variables = None) -> dict[str, Any]:
        """
        Execute a GraphQL query and return the raw GraphQL response as a python dictionary.

        Usage:
            # Example 1:
            data = st.graphql.raw_query("query rules($first: Int) { rules { values(first: $first, offset: 0) { id } } }", variables={"first": 100})

        """
        json = (
            {"query": query, "variables": variables} if variables else {"query": query}
        )

        response = self.session.post("graphql", json=json)
        response.raise_for_status()

        res = response.json()
        if "errors" in res:
            raise GraphqlError(errors=res["errors"])

        return response.json()



from datetime import datetime
from enum import Enum

from typing import List, Optional, Union

from pytos2.api import SdkError
from pytos2.secureapp.network_object import NetworkObject, classify_network_object
from pytos2.secureapp.service_object import Service, classify_service_object
from pytos2.securechange.user import SCWUser

from ..utils import propify, prop, safe_iso8601_date, kwargify
from ..models import Jsonable, Link, ObjectReference


@propify
class ApplicationConnection(Jsonable):
    class Meta(Enum):
        ROOT = "connection"
        ENTITY = "application"

    id: Optional[int] = prop(None)
    services: List[ObjectReference] = prop(flatify="service", factory=list)
    sources: List[ObjectReference] = prop(flatify="source", factory=list)
    destinations: List[ObjectReference] = prop(flatify="destination", factory=list)
    status: Optional[str] = prop(None)
    external: Optional[bool] = prop(None)
    connection_to_application: Optional[ObjectReference] = prop(None)
    comment: Optional[str] = prop(None)
    uid: Optional[str] = prop(None)
    application_id: Optional[str] = prop(None, key="applicationId")
    open_tickets: List[ObjectReference] = prop(flatify="ticket", factory=list)
    name: Optional[str] = prop(None)

    def delete(self):
        from pytos2.secureapp.entrypoint import Sa

        Sa.default.delete_application_connection(int(self.application_id), self.id)

    def update(self):
        if not self.application_id or not self.id:
            raise SdkError(
                "application_id and id are required to update application connection."
            )

        from pytos2.secureapp.entrypoint import Sa

        return Sa.default.update_application_connection(int(self.application_id), self)


@propify
class ExtendedApplicationConnection(ApplicationConnection):
    id: Optional[int] = prop(None)
    services: List[Service] = prop(
        flatify="service", factory=list, kwargify=classify_service_object
    )
    sources: List[NetworkObject] = prop(
        flatify="source", factory=list, kwargify=classify_network_object
    )
    destinations: List[NetworkObject] = prop(
        flatify="destination", factory=list, kwargify=classify_network_object
    )


@propify
class InterfaceConnection(Jsonable):
    class Meta(Enum):
        ROOT = "interface_connection"
        ENTITY = "interface connection"

    name: Optional[str] = prop(None)
    comment: Optional[str] = prop(None)
    open_tickets: List[ObjectReference] = prop(flatify="ticket", factory=list)
    services: List[ObjectReference] = prop(flatify="service", factory=list)
    sources: List[ObjectReference] = prop(flatify="source", factory=list)
    destinations: List[ObjectReference] = prop(flatify="destination", factory=list)
    connected_servers: List[ObjectReference] = prop(flatify="server", factory=list)


@propify
class ApplicationInterface(Jsonable):
    class Meta(Enum):
        ROOT = "application_interface"
        ENTITY = "application interface"

    name: Optional[str] = prop(None)
    comment: Optional[str] = prop(None)
    is_published: Optional[bool] = prop(None)
    application_id: Optional[int] = prop(None)
    interface_connections: List[InterfaceConnection] = prop(
        flatify="interface_connection", factory=list
    )


@propify
class Connection(Jsonable):
    class Prop(Enum):
        APPLICATION_ID = "applicationId"

    class Meta(Enum):
        ROOT = "connection"
        ENTITY = "connection"

    id: Optional[str | int] = prop(None)
    services: List[ObjectReference] = prop(flatify="service", factory=list)
    sources: List[ObjectReference] = prop(flatify="source", factory=list)
    destinations: List[ObjectReference] = prop(flatify="destination", factory=list)
    status: Optional[str] = prop(None)
    external: Optional[bool] = prop(None)
    connection_to_application: Optional[ObjectReference] = prop(None)
    comment: Optional[str] = prop(None)
    name: str = prop()
    uid: Optional[str] = prop(None)
    application_id: Optional[str] = prop(None, key=Prop.APPLICATION_ID.value)
    open_tickets: List[ObjectReference] = prop(flatify="ticket", factory=list)


@propify
class ApplicationInterfaceInstanceGet(Jsonable):
    class Meta(Enum):
        ROOT = "connection_to_application"
        ENTITY = "connection to application"

    id: Optional[str | int] = prop(None)
    uid: Optional[str] = prop(None)
    name: Optional[str] = prop(None)
    comment: Optional[str] = prop(None)
    application_id: Optional[str | int] = prop(None)
    connections: List[Connection] = prop(factory=list)
    application_interface_id: Optional[str | int] = prop(None)


@propify
class ApplicationInterfaceInstance(Jsonable):
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)
    servers: List[ObjectReference] = prop(flatify="server", factory=list)
    application_interface_id: Optional[str | int] = prop(None)


@propify
class Application(Jsonable):
    class Meta(Enum):
        ROOT = "application"
        ENTITY = "application"

    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    customer: Optional[ObjectReference] = prop(None)
    status: Optional[str] = prop(None)
    vendors: List[str] = prop(flatify="vendor", factory=list, jsonify=False)
    created: Optional[datetime] = prop(None, kwargify=safe_iso8601_date, jsonify=False)
    connections: List[ObjectReference] = prop(
        flatify="connection", factory=list, jsonify=False
    )
    decommissioned: Optional[bool] = prop(None)
    editors: List[ObjectReference] = prop(factory=list)
    viewers: List[ObjectReference] = prop(factory=list)
    open_tickets: List[ObjectReference] = prop(factory=list, flatify="ticket")
    connection_to_application_packs: List[ObjectReference] = prop(
        factory=list, flatify="connection_to_application_pack", jsonify=False
    )
    name: str = prop()
    owner: Optional[ObjectReference] = prop(None)
    modified: Optional[datetime] = prop(None, kwargify=safe_iso8601_date, jsonify=False)

    def delete(self):
        from pytos2.secureapp.entrypoint import Sa

        Sa.default.delete_application(self.id)

    def update(self):
        from pytos2.secureapp.entrypoint import Sa

        if not self.id:
            raise SdkError("id is required to update application.")

        return Sa.default.update_application(self)



import pytest
import json
from typing import List
from netaddr import IPAddress
import responses
from . import conftest

from pytos2.api import Pager
from pytos2.securetrack.topology import (
    TopologyMode,
    TopologySyncStatus,
    TopologySubnet,
    TopologySubnetDetailed,
    SubnetDevice,
)


class TestTopology:
    @responses.activate
    def test_topology_subnets(self, topology_subnets_mock, st):
        pager: Pager = st.get_topology_subnets()
        subnets: List[TopologySubnet] = pager.fetch_all()
        for sub in subnets:
            assert isinstance(sub.id, int)
            assert isinstance(sub.name, str)
            assert isinstance(sub.ip, IPAddress)
            assert isinstance(sub.mask, IPAddress)
            if sub.domain_id_tag:
                assert isinstance(sub.domain_id_tag, int)

        sub = subnets[0]
        assert sub.id == 6
        assert sub.name == "10.1.1.0/30"
        assert sub.ip == IPAddress("10.1.1.0")
        assert sub.mask == IPAddress("255.255.255.252")
        assert sub.domain_id_tag == 1

        sub: TopologySubnetDetailed = st.get_topology_subnet(6)
        assert sub.id == 6
        assert sub.name == "10.1.1.0/30"
        assert sub.ip == IPAddress("10.1.1.0")
        assert sub.mask == IPAddress("255.255.255.252")
        if sub.domain_id_tag:
            assert sub.domain_id_tag == 1

        device = sub.attached_devices[0]
        assert isinstance(device, SubnetDevice)
        assert device.id == 1
        assert device.is_generic is False
        assert device.name == "RTR1"
        assert device.interface_id == 1
        assert device.interface_name == "GigabitEthernet2"
        assert device.interface_ip == IPAddress("10.1.1.2")

        assert isinstance(sub.join_candidates, List)
